package com.wmp.nofrills3dtest;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import android.app.Activity;
import android.opengl.GLES20;
import android.os.Bundle;
import android.view.Menu;
import android.view.Window;
import android.view.WindowManager;

import com.wmp.nofrills3d.entities.Entity;
import com.wmp.nofrills3d.entities.Model;
import com.wmp.nofrills3d.entities.ProjectiveCamera;
import com.wmp.nofrills3d.entities.SkinnedModel;
import com.wmp.nofrills3d.entities.Transform;
import com.wmp.nofrills3d.entities.loaders.ModelLoader;
import com.wmp.nofrills3d.entities.loaders.ModelLoaderDelegate;
import com.wmp.nofrills3d.geometry.IMesh;
import com.wmp.nofrills3d.geometry.Mesh;
import com.wmp.nofrills3d.geometry.PrimitiveMeshFactory;
import com.wmp.nofrills3d.ik.IKConstraint;
import com.wmp.nofrills3d.input.UserInputEvent;
import com.wmp.nofrills3d.input.UserInputEvent.EventActionEnum;
import com.wmp.nofrills3d.material.Material;
import com.wmp.nofrills3d.material.Texture;
import com.wmp.nofrills3d.math.Matrix4;
import com.wmp.nofrills3d.math.Plane;
import com.wmp.nofrills3d.math.Ray;
import com.wmp.nofrills3d.math.Vector3;
import com.wmp.nofrills3d.physics.BoneCollider;
import com.wmp.nofrills3d.physics.CollisionDetector;
import com.wmp.nofrills3d.physics.Contact;
import com.wmp.nofrills3d.physics.ContactCollection;
import com.wmp.nofrills3d.physics.ICollider;
import com.wmp.nofrills3d.physics.SphereCollider;
import com.wmp.nofrills3d.scene.ISceneDelegate;
import com.wmp.nofrills3d.scene.Scene;
import com.wmp.nofrills3d.scene.SceneView;
import com.wmp.nofrills3d.utils.ContentLoader;
import com.wmp.nofrills3d.utils.Logger;
import com.wmp.wmp3dtest.R;

public class MeshTestActivity extends Activity implements ISceneDelegate, ModelLoaderDelegate {

	public static final String TAG = "MeshTestActivity";		
	
	public static final float GROUND_Y = -5.0f;
	
	private static final boolean RENDER_COLLIDERS = false; 
	
	private static final boolean RENDER_BONES = false; 
	
	private static final boolean RENDER_CHAINS = false; 
	
	protected ProjectiveCamera camera;
	protected SceneView view; 	
	
	private Model sphereModel = null;	
	
	private Transform draggedTransform = null; 
	
	private Vector3 draggedTransformOffset = Vector3.createZero(); 
	
	private Model bananaModel = null; 
	
	private Monkey monkey = null;
	
	protected Hashtable<String, ModelLoader> loadedModels = new Hashtable<String, ModelLoader>();  
	
	@Override
	protected void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);
		
		requestWindowFeature(Window.FEATURE_NO_TITLE);
        getWindow().addFlags(WindowManager.LayoutParams.FLAG_FULLSCREEN);
        getWindow().addFlags(WindowManager.LayoutParams.FLAG_KEEP_SCREEN_ON);
        
        // init view 
		camera = new ProjectiveCamera();
		camera.move(0.0f, 2.0f, 20.0f);
		//camera.rotate(5.0f, 0.0f, 0.0f);
		camera.lookAt(new Vector3(0.0f, 0.0f, 0.0f));
		view = new SceneView(this);
		view.setCamera(camera);
		view.setDelegate(this);
		
		setContentView(view);
		
		loadModels(); 
	}

	@Override
	public boolean onCreateOptionsMenu(Menu menu) {		
		return true;
	}
	
	@Override
	protected void onResume(){
		super.onResume();
		
		if( view != null ){
			view.onResume();
		}
	}
	
	@Override
	protected void onPause(){
		super.onPause();
		
		if( view != null ){
			view.onPause();
		}
	}			
	
	private void loadModels(){
		loadAsyncModel("ground", ContentLoader.ModelTypeEnum.dae, R.raw.ground);
		loadAsyncModel("cloud", ContentLoader.ModelTypeEnum.dae, R.raw.cloud);
		loadAsyncModel("tree", ContentLoader.ModelTypeEnum.dae, R.raw.tree);
		loadAsyncModel("monkey", ContentLoader.ModelTypeEnum.dae, R.raw.monkey_v25);
		loadAsyncModel("banana", ContentLoader.ModelTypeEnum.dae, R.raw.banana);
	}

	/* (non-Javadoc)
	 * @see com.wmp.wmp3d.game.GameDelegate#onSurfaceCreated()
	 */
	@Override
	public void onSurfaceCreated(Scene scene) {		
		
		Mesh mesh = PrimitiveMeshFactory.createSphere(1.0f, 12, 12);
		mesh.glDrawingMode = GLES20.GL_LINES;
		Material mat = new Material(this, R.raw.basic_texture_vs, R.raw.basic_texture_fs);
		sphereModel = new Model(mesh, mat);
		sphereModel.visible = false; 
		sphereModel.setPosition(0.0f, 0.0f, -10.0f);		
		
		scene.addEntity(sphereModel);
		
		if( loadedModels.size() > 0 ){
			Enumeration<ModelLoader> enu = loadedModels.elements();
			while(enu.hasMoreElements() ){
				ModelLoader loader = enu.nextElement();
				addEntityWithModelLoader(loader);
			}
		}
		
	}
	
	private void loadAsyncModel(String tag, ContentLoader.ModelTypeEnum modelType, int... resId){
		ModelLoader modelLoader = ContentLoader.createModelLoader(this, modelType);
		modelLoader.tag = tag;
		modelLoader.parseAsync(this, resId);
	}
	
	private void addEntityWithModelLoader(ModelLoader loader){
		loadedModels.remove(loader);
		
		if( loader.tag.equals("ground") ){
			addGround(loader);
		}	
		else if( loader.tag.equals("tree") ){
			addTrees(loader);
		}
		else if( loader.tag.equals("cloud") ){
			addClouds(loader);
		}
		else if( loader.tag.equals("monkey") ){
			addMonkey(loader);
		}
		else if( loader.tag.equals("banana") ){
			addBanana(loader);
		}
	}
	
	private void addGround(ModelLoader modelLoader){
		Texture tex = ContentLoader.loadTextureFromResource(this, R.raw.ground_tex);				
		Material mat = (Material)modelLoader.getDefaultMaterial();
		IMesh[] loadedMeshes = modelLoader.buildAndFetchGeometry(); 				

		Model model = new Model(loadedMeshes[0], mat, tex);
		model.setPosition(0.0f, GROUND_Y, 0.0f);
		model.setRotation(-90.0f, 0.0f, 0.0f); 
		view.addEntity(model);		
	}
	
	private void addClouds(ModelLoader modelLoader){			
		IMesh[] loadedMeshes = modelLoader.buildAndFetchGeometry(); 
		Material mat = (Material)modelLoader.getDefaultMaterial();
		Texture tex = ContentLoader.loadTextureFromResource(this, R.raw.cloud_tex);

		Model model = new Model(loadedMeshes[0], mat, tex);
		model.setPosition(0.0f, 5.0f, -10.0f);
		model.setRotation(-90.0f, 0.0f, 0.0f);
		model.setScale(1.1f);
		view.addEntity(model);
		
		model = new Model(loadedMeshes[0], mat, tex);
		model.setPosition(2.0f, 7.0f, -10.0f);
		model.setRotation(-90.0f, 0.0f, 0.0f);
		model.setScale(0.9f);
		view.addEntity(model);
	}
	
	private void addTrees(ModelLoader modelLoader){
		Texture tex = ContentLoader.loadTextureFromResource(this, R.raw.tree_tex);		
		Material mat = (Material)modelLoader.getDefaultMaterial(); 
		IMesh[] loadedMeshes = modelLoader.buildAndFetchGeometry(); 
		
		final float treeOffsetY = -0.2f;

		Model model = new Model(loadedMeshes[0], mat, tex);
		model.setPosition(5.0f, GROUND_Y + treeOffsetY, -5.0f);
		model.setRotation(-90.0f, 0.0f, 0.0f); 
		view.addEntity(model);
		
		model = new Model(loadedMeshes[0], mat, tex);
		model.setPosition(2.0f, GROUND_Y + treeOffsetY, -7.0f);
		model.setRotation(-90.0f, 0.0f, 0.0f); 
		view.addEntity(model);
		
		model = new Model(loadedMeshes[0], mat, tex);
		model.setPosition(-3.0f, GROUND_Y + treeOffsetY, 5.0f);
		model.setRotation(-90.0f, 0.0f, 0.0f); 
		view.addEntity(model);
	}
	
	private void addMonkey(ModelLoader modelLoader){
		monkey = new Monkey(modelLoader);
		monkey.setPosition(0.0f, GROUND_Y, 0.0f); 
		view.addEntity(monkey);	
	}
	
	private void addBanana(ModelLoader modelLoader){
		Texture tex = ContentLoader.loadTextureFromResource(this, R.raw.banana_tex);		
		Material mat = (Material)modelLoader.getDefaultMaterial(); 
		IMesh[] loadedMeshes = modelLoader.buildAndFetchGeometry(); 
		
		bananaModel = new Model(loadedMeshes[0], mat, tex);
		bananaModel.setPosition(2.0f, GROUND_Y + 2.0f, 1.5f);
		bananaModel.setRotation(-90.0f, 0.0f, 0.0f);
		bananaModel.tag = "banana";
		view.addEntity(bananaModel);	
		
		SphereCollider collider = new SphereCollider(bananaModel);
		collider.radius = 0.5f;
		bananaModel.addCollider(collider);
	}

	/* (non-Javadoc)
	 * @see com.wmp.wmp3d.game.GameDelegate#unlockContent()
	 */
	@Override
	public void unloadScene(Scene scene) {
		
		
	}

	@Override
	public void onPostUpdate(Scene scene, float elapsedTime) {
		
		//camera.move(0.0f, 0.0f, elapsedTime * 1.0f);
		//camera.rotate(0.0f, elapsedTime * -50.0f, 0.0f);
		
		if( scene.hasUserInputEvents() ){
			Iterator<UserInputEvent> it = scene.getUserInputEventsIterator(); 
			while( it.hasNext() ){
				UserInputEvent event = it.next();
				
				if( event.action == EventActionEnum.Down ){
					Ray ray = scene.screenPositionToRay(event.x, event.y);
					Transform transform = findTouchedEntity(ray, scene);
					
					if( transform != null ){
						Logger.Log(TAG, "TOUCHED!!!! " + transform.toString() );
					
						if( transform == bananaModel ){ 
							draggedTransform = transform; 
							draggedTransformOffset.set(0, 0, 0); 
							monkey.setTarget(draggedTransform.getPosition()); 
						}
						
//						if( transform instanceof BoneCollider ){
//							BoneCollider bc = ((BoneCollider)transform);
//							
//							if( bc.getBone().id.equalsIgnoreCase("hand_L") ){
//								SkinnedModel sm = (SkinnedModel)bc.getTransform();
//								IKChain ikChain = new IKChain(sm, bc.getBone(), 2);
//								sm.attachIKChain(ikChain);
//								ikChain.activate(); 
//								ikChain.setTarget(Vector3.add(sm.getPosition(), new Vector3(2.5f, 0.0f, 0.0f)));
//								
//								//bc.getBone().setLock(true, 3); 
//							} 
//							
//							if( bc.getBone().id.equalsIgnoreCase("head") ){
//								Transform parentTransform = bc.getTransform(); 
//								draggedTransform = parentTransform; 
//								draggedTransformOffset = Vector3.subtract(transform.getPosition(), parentTransform.getPosition()); 
//							}														
//						}						
					} 
				}
				else if( event.action == EventActionEnum.Moved ){
					
					if( draggedTransform != null ){
						Ray ray = scene.screenPositionToRay(event.x, event.y);
						
			            Plane plane = new Plane(new Vector3(0, 0, draggedTransform.getPosition().z), new Vector3(0, 0, 1));
	
			            // Find out where the touched point intersects the plane
			            // representing our table. We'll move the mallet along this plane.
			            Vector3 touchedPoint = CollisionDetector.getIntersectionPoint(ray, plane);
			            
			            draggedTransform.setPosition(touchedPoint.x + draggedTransformOffset.x, touchedPoint.y + draggedTransformOffset.y, draggedTransform.getPosition().z);
			            monkey.updateTarget(draggedTransform.getPosition());
					}
				}
				else if( event.action == EventActionEnum.Up ){
					if( draggedTransform != null ){
						draggedTransform = null;
						monkey.releaseTarget(); 
					}
				}
			}
		}
	}
	
	public Transform findTouchedEntity(Ray ray, Scene scene){
		Iterator<Entity> it = scene.getEntityIterator();
		
		ContactCollection contactCollection = new ContactCollection();
				
		while( it.hasNext() ){
			Entity entity = it.next(); 
			
			if( !entity.visible || !entity.enabled ){
				continue; 
			}
			
			Iterator<ICollider> colliderIt = entity.getColliderCollectionIterator();
			while( colliderIt.hasNext() ){
				ICollider c = colliderIt.next();
				
				if( ray.collidesWith(c, contactCollection) > 0 ){
					
				}
			}
		}
		
		contactCollection.sort(); 
		
		if( contactCollection.size() > 0 ){
			Contact contact = contactCollection.getContact(0);
			if( contact.colliderA == ray ){
				return contact.colliderB.getTransform(); 
			} else{
				return contact.colliderA.getTransform(); 
			} 
		}
		
		return null; 
	}

	@Override
	public void onPostRender(Scene scene) {
		if( RENDER_COLLIDERS ){
			drawColliders(scene);
		}
		
		if( RENDER_BONES ){
			drawBones(scene);
		}
		
		if( RENDER_CHAINS ){
			drawIKChains(scene);
		}
	}
	
	protected void drawBones(Scene scene){
		Iterator<Entity> it = scene.getEntityIterator();
		
		while( it.hasNext() ){
			Entity entity = it.next(); 
			
			if( !entity.visible ){
				//continue; 
			}
			
			if( entity instanceof SkinnedModel ){
				Vector3 tmp = new Vector3(); 
				
				SkinnedModel sm = (SkinnedModel)entity; 
				
				for( int i=0; i<sm.worldTransforms.length; i++ ){
					Matrix4 matrix = sm.worldTransforms[i];
					Matrix4.multiply(sm.modelTransform, matrix, matrix);
					Matrix4.decomposeTranslation(matrix, tmp);
					
					sphereModel.setScale(0.2f);
					sphereModel.setPosition(tmp);
					sphereModel.update(0);
					sphereModel.render(camera.projectionMatrix, camera.modelTransform);
				}
			}
		}
	}
	
	protected void drawColliders(Scene scene){
		Iterator<Entity> it = scene.getEntityIterator();
		
		while( it.hasNext() ){
			Entity entity = it.next(); 
			
			if( !entity.visible ){
				//continue; 
			}
			
			Iterator<ICollider> colliderIt = entity.getColliderCollectionIterator();
			while( colliderIt.hasNext() ){
				ICollider c = colliderIt.next(); 
				
				if( c instanceof BoneCollider ){
					BoneCollider sc = (BoneCollider)c;
					if( sc.getCollider().getColliderType() == ICollider.COLLIDER_TYPE_SPHERE ){
						sphereModel.setScale(((SphereCollider)sc.getCollider()).radius);
					} else{
						sphereModel.setScale(1.0f);
					}					
					sphereModel.setPosition(sc.getPosition());
					sphereModel.update(0);
					sphereModel.render(camera.projectionMatrix, camera.modelTransform);
				} else{
					if( c.getColliderType() == ICollider.COLLIDER_TYPE_SPHERE ){
						sphereModel.setScale(((SphereCollider)c).radius);
					} else{
						sphereModel.setScale(1.0f);
					}					
					sphereModel.setPosition(((SphereCollider)c).getCentre());
					sphereModel.update(0);
					sphereModel.render(camera.projectionMatrix, camera.modelTransform);
				}
			}
		}
	}
	
	protected void drawIKChains(Scene scene){
		Iterator<Entity> it = scene.getEntityIterator();
		
		while( it.hasNext() ){
			Entity entity = it.next(); 
			
			if( !entity.visible ){
				continue; 
			}
			
			if( entity instanceof SkinnedModel ){
				SkinnedModel sm = (SkinnedModel)entity; 
				
				Iterator<IKConstraint> itB = sm.getIKChainsIterator(); 
				while( itB.hasNext() ){
					IKConstraint ikChain = itB.next(); 
					
					if( ikChain.isActive() ){
						Vector3 targetPosition = ikChain.getTarget();
						sphereModel.setScale(0.5f);
						sphereModel.setPosition(targetPosition);
						sphereModel.update(0);
						sphereModel.render(camera.projectionMatrix, camera.modelTransform);
					}
				}
			}												
		}
	}

	@Override
	public void onModelResourcesParsed(final ModelLoader loader, int state) {
		// remove from queue 
		loadedModels.put(loader.tag, loader); 
		
		if( !view.isGLInitilised() ){
			return; 
		}
		
		// run on the GL thread 
		this.view.queueEvent(new Runnable() {
			@Override
			public void run() {		
				addEntityWithModelLoader(loader);				
			}
		});				
	}
}
