package com.wmp.nofrills3dtest;

import android.content.Context;

import com.wmp.nofrills3d.animations.AnimationClip;
import com.wmp.nofrills3d.animations.AnimationClipSequence;
import com.wmp.nofrills3d.entities.SkinnedModel;
import com.wmp.nofrills3d.entities.loaders.ModelLoader;
import com.wmp.nofrills3d.entities.loaders.ModelLoaderDelegate;
import com.wmp.nofrills3d.geometry.IMesh;
import com.wmp.nofrills3d.geometry.SkinnedMesh;
import com.wmp.nofrills3d.ik.IKChain;
import com.wmp.nofrills3d.ik.IKConstraint;
import com.wmp.nofrills3d.ik.IKConstraintDelegate;
import com.wmp.nofrills3d.ik.IKPivot;
import com.wmp.nofrills3d.material.Texture;
import com.wmp.nofrills3d.material.SkinnedMaterial;
import com.wmp.nofrills3d.math.Vector3;
import com.wmp.nofrills3d.physics.SphereCollider;
import com.wmp.nofrills3d.utils.ContentLoader;
import com.wmp.nofrills3d.utils.Logger;
import com.wmp.wmp3dtest.R;

public class Monkey extends SkinnedModel implements IKConstraintDelegate{

	public static final String TAG = "Monkey";
	
	private static final int IK_LEFTHAND 	= 100; 
	private static final int IK_RIGHTHAND 	= 150;
	private static final int IK_HEAD 	 	= 200;
	
	protected Vector3 target = Vector3.createZero();
	
	protected boolean targetSet = false; 
	
	public enum SideEnum{
		Left, 
		Right
	}
	
	public enum StateEnum{
		Idle, 
		Walking 
	}
	
	protected StateEnum state = StateEnum.Idle; 
	
	protected float stateElapsedTime = 0.0f; 
	
	public Monkey(ModelLoader loader){		
		super(); 						
		
		IMesh[] loadedMeshes = loader.buildAndFetchGeometry(); 
		
		Texture tex = ContentLoader.loadTextureFromResource(loader.getContext(), R.raw.monkeytex);
		SkinnedMaterial mat = (SkinnedMaterial) loader.getDefaultMaterial();
		
		if( loadedMeshes != null && loadedMeshes.length == 1 ){
			initModel(loadedMeshes[0], mat, tex);
		}				
		
		SphereCollider headCollider = new SphereCollider(this);
		headCollider.radius = 0.8f;
		headCollider.centre.y = 0.3f; 
		attachColliderToBone(findBoneWithId("head"), headCollider);
		
		SphereCollider leftHandCollider = new SphereCollider(this);
		leftHandCollider.radius = 0.3f;
		leftHandCollider.centre.y = 0.4f;
		attachColliderToBone(findBoneWithId("hand_L"), leftHandCollider);
		
		SphereCollider rightHandCollider = new SphereCollider(this);
		rightHandCollider.radius = 0.3f;
		rightHandCollider.centre.y = 0.4f; 
		attachColliderToBone(findBoneWithId("hand_R"), rightHandCollider);
		
		initIKConstraints();
	}
	
	public Monkey(SkinnedMesh mesh, SkinnedMaterial material) {
		super(mesh, material);
		// TODO Auto-generated constructor stub
	}
	
	@Override
	public void update(float elapsedTime){
		
		//this.rotation.y += 10.0 * elapsedTime;
		//this.position.add(modelTransform.getForward().scale(1.0f * elapsedTime));
		
		super.update(elapsedTime);
		
		if( this.enabled ){
			stateElapsedTime += elapsedTime; 
			
			if( state == StateEnum.Idle ){
				if( stateElapsedTime >= animationPlayer.getTimeRemaining() - 0.5f ){
					setState(StateEnum.Walking);
				}
			} 
			
			else if( state == StateEnum.Walking ){
//				Logger.Log(TAG, modelTransform.getForward().toString() 
//						+ ", " + modelTransform.getUp().toString()
//							+ ", " + modelTransform.getRight().toString());
			}
		}
		
		//Logger.Log("RIGHT", this.getRight().toString());
		//Logger.Log("UP", this.getUp().toString());
		//Logger.Log("FORWARD", this.getForward().toString());
		
		//position.add(this.getRight().scale(elapsedTime * 2)); 
		//rotation.y += elapsedTime * 60.0f; 
	}	
	
	public void setState(StateEnum newState){
		if( state == newState ){
			return; 
		}
		
		state = newState; 
		stateElapsedTime = 0.0f;
		
		if( state == StateEnum.Idle ){
			
		}
		else if( state == StateEnum.Walking ){
			//animationPlayer.crossFadeAnimation(animationSequences.get("walk"), animationPlayer.getTimeRemaining());
			//animationPlayer.crossFadeAnimation(animationSequences.get("walk"), 0.5f);
			//animationPlayer.startAnimation(animationSequences.get("walk"));
		}
	}
	
	@Override
	public void load(Context context){
		super.load(context);									
	}
	
	protected void initIKConstraints(){
		IKChain leftHandIKChain = new IKChain(this, findBoneWithId("hand_L"), 2);
		leftHandIKChain.id = IK_LEFTHAND;
		leftHandIKChain.delegate = this; 
		attachIKChain(leftHandIKChain); 
		
		IKChain rightHandIKChain = new IKChain(this, findBoneWithId("hand_R"), 2);
		rightHandIKChain.id = IK_RIGHTHAND; 
		rightHandIKChain.delegate = this; 
		attachIKChain(rightHandIKChain); 
		
		IKPivot headIK = new IKPivot(this, findBoneWithId("head")); 
		headIK.id = IK_HEAD;
		headIK.ikSpeed = 0.05f; 
		headIK.delegate = this; 
		attachIKChain(headIK);
	}
	
	@Override
	protected void initAnimationClipSequences(SkinnedMesh skinnedMesh){
		AnimationClip animationClip = skinnedMesh.animationClips.get(0); 
		animationClip.init(skinnedMesh.rootBone); 
		
		AnimationClipSequence defaultAnimation = new AnimationClipSequence(animationClip, 1.0f/24.0f, 35.0f/24.0f);
		defaultAnimation.speed = 1.0f; 
		defaultAnimation.tag = "default";
		defaultAnimation.looping = true; 
		animationSequences.put("default", defaultAnimation);
		
		AnimationClipSequence walkAnim = new AnimationClipSequence(animationClip, 40.0f/24.0f, 50.0f/24.0f);
		walkAnim.tag = "walk";
		walkAnim.looping = true; 
		animationSequences.put("walk", walkAnim);
		
		animationPlayer.startAnimation(defaultAnimation);
	}
	
	@Override
	public void onAnimationFinished(AnimationClipSequence animationClip) {		
		
	}

	@Override
	public void onTargetReached(IKConstraint ikConstraint) {
		// TODO Auto-generated method stub
		
	}
	
	public void setTarget(Vector3 targetWorldPosition){
		targetSet =  true; 
		target.set(targetWorldPosition);
		
		if( getTargetSide(targetWorldPosition) == SideEnum.Left ){
			findIKChainWithId(IK_LEFTHAND).setTarget(target, true);
			findIKChainWithId(IK_LEFTHAND).activate();
		} else{
			findIKChainWithId(IK_RIGHTHAND).setTarget(target, true);
			findIKChainWithId(IK_RIGHTHAND).activate();
		}
		
		findIKChainWithId(IK_HEAD).setTarget(target, true);
		findIKChainWithId(IK_HEAD).activate();				
	}
	
	public void updateTarget(Vector3 targetWorldPosition){
		targetSet =  true; 
		target.set(targetWorldPosition);
		
		if( getTargetSide(targetWorldPosition) == SideEnum.Left ){
			findIKChainWithId(IK_LEFTHAND).setTarget(target, true);
			findIKChainWithId(IK_LEFTHAND).activate();
			
			if( findIKChainWithId(IK_RIGHTHAND).isActive() ){
				findIKChainWithId(IK_RIGHTHAND).deactivate();
			}
		} else{
			findIKChainWithId(IK_RIGHTHAND).setTarget(target, true);
			findIKChainWithId(IK_RIGHTHAND).activate();
			
			if( findIKChainWithId(IK_LEFTHAND).isActive() ){
				findIKChainWithId(IK_LEFTHAND).deactivate();
			}
		}
		
		if(findIKChainWithId(IK_HEAD).isActive() ){			
			findIKChainWithId(IK_HEAD).setTarget(target, true);
			
			findIKChainWithId(IK_LEFTHAND).setTarget(target, true);
		}
	}
	
	public void releaseTarget(){
		targetSet = false; 
		
		if( findIKChainWithId(IK_LEFTHAND).isActive() ){
			findIKChainWithId(IK_LEFTHAND).deactivate();
		}
		
		if( findIKChainWithId(IK_RIGHTHAND).isActive() ){
			findIKChainWithId(IK_RIGHTHAND).deactivate();
		}
		
		if( findIKChainWithId(IK_HEAD).isActive() ){
			findIKChainWithId(IK_HEAD).deactivate();
		}
	}
	
	protected SideEnum getTargetSide(Vector3 targetPosition){
		Vector3 dirToTarget = Vector3.subtract(this.position, targetPosition).normalise(); 
		float dot = Vector3.dotProduct(this.getRight(), dirToTarget);
		
		return dot < 0 ? SideEnum.Left : SideEnum.Right; 
	}
}
