package com.wmp.nofrills3dtest.wallpaper;

import java.util.Enumeration;
import java.util.Hashtable;

import android.app.ActivityManager;
import android.content.Context;
import android.content.pm.ConfigurationInfo;
import android.opengl.GLSurfaceView;
import android.os.Build;
import android.service.wallpaper.WallpaperService;
import android.view.SurfaceHolder;

import com.wmp.nofrills3d.entities.Camera;
import com.wmp.nofrills3d.entities.Entity;
import com.wmp.nofrills3d.entities.Model;
import com.wmp.nofrills3d.entities.ProjectiveCamera;
import com.wmp.nofrills3d.entities.loaders.ModelLoader;
import com.wmp.nofrills3d.entities.loaders.ModelLoaderDelegate;
import com.wmp.nofrills3d.geometry.IMesh;
import com.wmp.nofrills3d.material.Material;
import com.wmp.nofrills3d.material.Texture;
import com.wmp.nofrills3d.math.Vector3;
import com.wmp.nofrills3d.physics.SphereCollider;
import com.wmp.nofrills3d.scene.ISceneDelegate;
import com.wmp.nofrills3d.scene.Scene;
import com.wmp.nofrills3d.scene.SceneRenderer;
import com.wmp.nofrills3d.utils.ContentLoader;
import com.wmp.nofrills3d.utils.Logger;
import com.wmp.nofrills3dtest.Monkey;
import com.wmp.wmp3dtest.R;

public class NoFrills3DWallpaper extends WallpaperService{

	@Override
	public Engine onCreateEngine() {
		return new NoFrills3DWallpaperEngine();
	}
	
	public class NoFrills3DWallpaperEngine extends Engine implements ISceneDelegate, ModelLoaderDelegate{
		public static final String TAG = "NoFrills3DWallpaperEngine";		
		
		public static final float GROUND_Y = -5.0f;		
		
		protected ProjectiveCamera camera;
		protected WallpaperSceneView view; 	
		
		protected Hashtable<String, ModelLoader> loadedModels = new Hashtable<String, ModelLoader>(); 
						
		
		private Model bananaModel = null; 
		
		private Monkey monkey = null;
		
		@Override
        public void onCreate(SurfaceHolder surfaceHolder) {
            super.onCreate(surfaceHolder);
            
            // init view 
			camera = new ProjectiveCamera();
			camera.move(0.0f, 2.0f, 20.0f);
			camera.lookAt(new Vector3(0.0f, 0.0f, 0.0f));			
			view = new WallpaperSceneView(NoFrills3DWallpaper.this);
			view.setCamera(camera);
			view.setDelegate(this);
    		
    		loadModels(); 
		}				
		
		@Override
        public void onVisibilityChanged(boolean visible) {
            super.onVisibilityChanged(visible);
            
            
            if (view.isGLInitilised()) {
                if (visible) {
                	view.onResume();
                } else {                   
                    view.onPause(); 
                }
            }
        }   
        
        @Override
        public void onOffsetsChanged(final float xOffset, final float yOffset,
            float xOffsetStep, float yOffsetStep, int xPixelOffset, int yPixelOffset) {
        	view.queueEvent(new Runnable() {                
                @Override
                public void run() {
                    //particlesRenderer.handleOffsetsChanged(xOffset, yOffset); 
                	
                	// move camera? 
                	
                }
            });            
        }
        @Override
        public void onDestroy() {
            super.onDestroy();
            view.onDestory(); 
        }

		@Override
		public void onModelResourcesParsed(final ModelLoader loader, int state) {
			// remove from queue 
			loadedModels.put(loader.tag, loader); 
			
			if( !view.isGLInitilised() ){
				return; 
			}
			
			// run on the GL thread 
			this.view.queueEvent(new Runnable() {
				@Override
				public void run() {		
					addEntityWithModelLoader(loader);				
				}
			});
			
		}

		@Override
		public void onSurfaceCreated(Scene scene) {
			if( loadedModels.size() > 0 ){
				Enumeration<ModelLoader> enu = loadedModels.elements();
				while(enu.hasMoreElements() ){
					ModelLoader loader = enu.nextElement();
					addEntityWithModelLoader(loader);
				}
			}			
		}

		@Override
		public void unloadScene(Scene scene) {
			// TODO Auto-generated method stub
			
		}

		@Override
		public void onPostUpdate(Scene scene, float elapsedTime) {
			// TODO Auto-generated method stub
			
		}

		@Override
		public void onPostRender(Scene scene) {
			// TODO Auto-generated method stub
			
		}
		
		private void loadModels(){
			loadAsyncModel("ground", ContentLoader.ModelTypeEnum.dae, R.raw.ground);
			loadAsyncModel("cloud", ContentLoader.ModelTypeEnum.dae, R.raw.cloud);
			loadAsyncModel("tree", ContentLoader.ModelTypeEnum.dae, R.raw.tree);
			//loadAsyncModel("monkey", ContentLoader.ModelTypeEnum.dae, R.raw.monkey_v25);
			loadAsyncModel("banana", ContentLoader.ModelTypeEnum.dae, R.raw.banana);
		}
		
		private void loadAsyncModel(String tag, ContentLoader.ModelTypeEnum modelType, int... resId){
			ModelLoader modelLoader = ContentLoader.createModelLoader(NoFrills3DWallpaper.this, modelType);
			modelLoader.tag = tag;
			modelLoader.parseAsync(this, resId);
		}
		
		private void addEntityWithModelLoader(ModelLoader loader){
			loadedModels.remove(loader);
			
			if( loader.tag.equals("ground") ){
				addGround(loader);
			}	
			else if( loader.tag.equals("tree") ){
				addTrees(loader);
			}
			else if( loader.tag.equals("cloud") ){
				addClouds(loader);
			}
			else if( loader.tag.equals("monkey") ){
				addMonkey(loader);
			}
			else if( loader.tag.equals("banana") ){
				addBanana(loader);
			}
		}
		
		private void addGround(ModelLoader modelLoader){
			Texture tex = ContentLoader.loadTextureFromResource(NoFrills3DWallpaper.this, R.raw.ground_tex);				
			Material mat = (Material)modelLoader.getDefaultMaterial();
			IMesh[] loadedMeshes = modelLoader.buildAndFetchGeometry(); 				

			Model model = new Model(loadedMeshes[0], mat, tex);
			model.setPosition(0.0f, GROUND_Y, 0.0f);
			model.setRotation(-90.0f, 0.0f, 0.0f); 
			view.addEntity(model);		
		}
		
		private void addClouds(ModelLoader modelLoader){			
			IMesh[] loadedMeshes = modelLoader.buildAndFetchGeometry(); 
			Material mat = (Material)modelLoader.getDefaultMaterial();
			Texture tex = ContentLoader.loadTextureFromResource(NoFrills3DWallpaper.this, R.raw.cloud_tex);

			Model model = new Model(loadedMeshes[0], mat, tex);
			model.setPosition(0.0f, 5.0f, -10.0f);
			model.setRotation(-90.0f, 0.0f, 0.0f);
			model.setScale(1.1f);
			view.addEntity(model);
			
			model = new Model(loadedMeshes[0], mat, tex);
			model.setPosition(2.0f, 7.0f, -10.0f);
			model.setRotation(-90.0f, 0.0f, 0.0f);
			model.setScale(0.9f);
			view.addEntity(model);
		}
		
		private void addTrees(ModelLoader modelLoader){
			Texture tex = ContentLoader.loadTextureFromResource(NoFrills3DWallpaper.this, R.raw.tree_tex);		
			Material mat = (Material)modelLoader.getDefaultMaterial(); 
			IMesh[] loadedMeshes = modelLoader.buildAndFetchGeometry(); 
			
			final float treeOffsetY = -0.2f;

			Model model = new Model(loadedMeshes[0], mat, tex);
			model.setPosition(5.0f, GROUND_Y + treeOffsetY, -5.0f);
			model.setRotation(-90.0f, 0.0f, 0.0f); 
			view.addEntity(model);
			
			model = new Model(loadedMeshes[0], mat, tex);
			model.setPosition(2.0f, GROUND_Y + treeOffsetY, -7.0f);
			model.setRotation(-90.0f, 0.0f, 0.0f); 
			view.addEntity(model);
			
			model = new Model(loadedMeshes[0], mat, tex);
			model.setPosition(-3.0f, GROUND_Y + treeOffsetY, 5.0f);
			model.setRotation(-90.0f, 0.0f, 0.0f); 
			view.addEntity(model);
		}
		
		private void addMonkey(ModelLoader modelLoader){
			//monkey = new Monkey(modelLoader);
			//monkey.setPosition(0.0f, GROUND_Y, 0.0f); 
			//view.addEntity(monkey);	
		}
		
		private void addBanana(ModelLoader modelLoader){
			Texture tex = ContentLoader.loadTextureFromResource(NoFrills3DWallpaper.this, R.raw.banana_tex);		
			Material mat = (Material)modelLoader.getDefaultMaterial(); 
			IMesh[] loadedMeshes = modelLoader.buildAndFetchGeometry(); 
			
			bananaModel = new Model(loadedMeshes[0], mat, tex);
			bananaModel.setPosition(2.0f, GROUND_Y + 2.0f, 1.5f);
			bananaModel.setRotation(-90.0f, 0.0f, 0.0f);
			bananaModel.tag = "banana";
			view.addEntity(bananaModel);	
			
			SphereCollider collider = new SphereCollider(bananaModel);
			collider.radius = 0.5f;
			bananaModel.addCollider(collider);
		}
		
		protected class WallpaperSceneView extends GLSurfaceView{

			protected SceneRenderer renderer; 					
			
			public WallpaperSceneView(Context context) {
				super(context);
				init(); 
			}
			
			@Override
            public SurfaceHolder getHolder() {
                return getSurfaceHolder();
            }				
			
			protected void init(){
				if( supportsOpenGL2() ){						
					this.setEGLContextClientVersion(2);
		            this.setRenderer(renderer = new SceneRenderer(this));
		            
		            this.setRenderMode(RENDERMODE_CONTINUOUSLY); 
				} else{
					// TODO handle exception 
				}
				
			}
			
			public boolean isGLInitilised(){
				return renderer.isGLInitilised(); 
			}
			
			public void onPause(){
				if( renderer != null ){
					renderer.onPause(); 
				}
				
				super.onPause(); 
			}
			
			public void onResume(){	
				super.onResume();
				
				if( renderer != null ){
					renderer.onResume(); 
				}					
			}	
			
			public void onDestory(){
				super.onDetachedFromWindow();
			}
			
			public void setFrameRate(long frameRate){
				this.renderer.setFramerate(frameRate);
			}	
			
			public void setCamera(Camera camera){
				renderer.setCamera(camera); 
			}
			
			public void addEntity(Entity entity){
				renderer.addEntity(entity);
			}
			
			public void removeEntity(Entity entity){
				renderer.removeEntity(entity);
			}
			
			public Entity findEntityWithTag(String tag){
				return renderer.findEntityWithTag(tag);
			}
			
			public void setDelegate(ISceneDelegate delegate){
				renderer.setDelegate(delegate); 		
			}
			
			public boolean supportsOpenGL2(){
				ActivityManager activityManager = (ActivityManager) getContext().getSystemService(Context.ACTIVITY_SERVICE);
		        ConfigurationInfo configurationInfo = activityManager.getDeviceConfigurationInfo();        
				return configurationInfo.reqGlEsVersion >= 0x20000
		                || (Build.VERSION.SDK_INT >= Build.VERSION_CODES.ICE_CREAM_SANDWICH_MR1
		                 && (Build.FINGERPRINT.startsWith("generic")
		                  || Build.FINGERPRINT.startsWith("unknown")
		                  || Build.MODEL.contains("google_sdk")
		                  || Build.MODEL.contains("Emulator")
		                  || Build.MODEL.contains("Android SDK built for x86")));
			}	
			
		}
	}		
}
